/*
** Daedalus (Version 2.5) File: threed.h
** By Walter D. Pullen, Astara@msn.com, http://www.astrolog.org/labyrnth.htm
**
** IMPORTANT NOTICE: Daedalus and all Maze generation and general
** graphics routines used in this program are Copyright (C) 1998-2013 by
** Walter D. Pullen. Permission is granted to freely use, modify, and
** distribute these routines provided these credits and notices remain
** unmodified with any altered or distributed versions of the program.
** The user does have all rights to Mazes and other graphic output
** they make in Daedalus, like a novel created in a word processor.
**
** More formally: This program is free software; you can redistribute it
** and/or modify it under the terms of the GNU General Public License as
** published by the Free Software Foundation; either version 2 of the
** License, or (at your option) any later version. This program is
** distributed in the hope that it will be useful and inspiring, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details, a copy of which is in the
** LICENSE.HTM included with Daedalus, and at http://www.gnu.org
**
** This file contains definitions for general graphics routines, that operate
** on a monochrome bitmap being treated as a 3D/4D bitmap, unrelated to Mazes.
**
** Created: 3/12/1990.
** Last code change: 10/10/2013.
*/

/*
******************************************************************************
** Constants
******************************************************************************
*/

#define xCubeMax xBitmap
#define yCubeMax yBitmap
#define zCubeMax xBitmap

#define DIRS3 6
#define DIRS4 8


/*
******************************************************************************
** Macros
******************************************************************************
*/

#define TX(t) ((t)->wClient1)
#define TY(t) ((t)->wClient2)
#define TZ(t) ((t)->wClient3)
#define TW(t) ((t)->wClient4)
#define Tx(t, x, z) (((z) % TW(t)) * TX(t) + (x))
#define Ty(t, y, z) (((z) / TW(t)) * TY(t) + (y))
#define TG(t, x, y, z) MG(t, Tx(t, x, z), Ty(t, y, z))
#define TP(t, x, y, z, o) MP(t, Tx(t, x, z), Ty(t, y, z), o)
#define T0(t, x, y, z) M0(t, Tx(t, x, z), Ty(t, y, z))
#define T1(t, x, y, z) M1(t, Tx(t, x, z), Ty(t, y, z))
#define TR(t, x, y, z) MR(t, Tx(t, x, z), Ty(t, y, z))
#define TxM(t, x) ((x) % TX(t))
#define TyM(t, y) ((y) % TY(t))
#define TzM(t, x, y) (((y) / TY(t)) * TW(t) + ((x) / TX(t)))

#define Ux(t, w, x) ((w) * TX(t) + (x))
#define Uy(t, y, z) ((z) * TY(t) + (y))
#define UG(t, w, x, y, z) MG(t, Ux(t, w, x), Uy(t, y, z))
#define UP(t, w, x, y, z, o) MP(t, Ux(t, w, x), Uy(t, y, z), o)
#define U0(t, w, x, y, z) M0(t, Ux(t, w, x), Uy(t, y, z))
#define U1(t, w, x, y, z) M1(t, Ux(t, w, x), Uy(t, y, z))
#define UR(t, w, x, y, z) MR(t, Ux(t, w, x), Uy(t, y, z))

#define CG3(c, x, y, z) CG(c, Tx(c, x, z), Ty(c, y, z))
#define CP3(c, x, y, z, kv) CP(c, Tx(c, x, z), Ty(c, y, z), kv)

#define DeallocateCube(t) DeallocateP(t)
#define F64K3(t) (TX(t) > 0xFFFF || TY(t) > 0xFFFF || TZ(t) > 0xFFFF)
#define FLegalCubeLevel(t, x, y) \
  ((x) >= 0 && (x) < TX(t) && (y) >= 0 && (y) < TY(t))
#define FLegalCube(t, x, y, z) \
  (FLegalCubeLevel(t, x, y) && (z) >= 0 && (z) < TZ(t))


/*
******************************************************************************
** 3D Bitmap Routines
******************************************************************************
*/

typedef bitmap cubemap;
typedef constbitmap constcubemap;

extern CONST int woff3[DIRS4 + 4], xoff3[DIRS4 + 4], yoff3[DIRS4 + 4],
  zoff3[DIRS4 + 4];

extern flag FAllocateCube(cubemap *, int, int, int, int);
extern flag FCubeSizeSet(cubemap *, int, int, int, int);
extern flag FTesseractSizeSet(cubemap *, int, int, int, int);
extern bit Tg(constcubemap, int, int, int);
extern void CubeBlock(cubemap, int, int, int, int, int, int, bit);
extern void CubeBlockC(colmap, int, int, int, int, int, int, KV);
extern int CubeFlipCore(cubemap, int, int, int, char);
extern void CubeFlip(cubemap *, colmap *, CONST char *);
extern void CubeFlip2(cubemap *, colmap *, int, int);
extern flag FCubeFill(cubemap, int, int, int, bit);
extern flag FReadCube(cubemap *, FILE *, int);
extern void WriteCube(constcubemap, FILE *, int, flag);

/* threed.h */
